// Добавление вспомогательной библиотеки с набором функций проводной связи
#include <SoftwareSerial.h>

// Распределение функций управления роботом на выводы ATMega328P платы Arduino Uno R3
// Сонар
#define pinEcho 2 // серый провод
#define pinTrig 4 // фиолетовый провод
// питание белый(VCC)-черный(GND)

// Драйвер моторов
#define MotorInput1 3 // Motor A-IB, черный продов
#define MotorInput2 9 // Motor A-IA, белый провод
#define MotorInput3 10 // Motor B-IB, фиолетовый провод
#define MotorInput4 11 // Motor B-IА, белый провод
// питание красный(VCC)-коричневый(GND)

// Трекеры слежения за полосой
#define RightTracker 8 // Оранжевый(S)-красный(VCC)-коричневый(GND)
#define LeftTracker 12 // Желтый(S)-зеленый(VCC)-синий(GND)

// Датчики препятствия
#define RightObstacle 6 // Зеленый(S)-фиолетовый(VCC)-синий(GND)
#define LeftObstacle 7 // Серый(S)-черный(VCC)-белый(GND)

// Селектор программы управления
#define ProgramSelect1 14 // Коричневый провод-перемычка
#define ProgramSelect2 15 // Синий провод-перемычка

// Программно формируемый канал UART для связи Arduino Uno с модулем Bluetooth
#define BT_RXpin 18 // Зеленый провод -> Bluetooth TX (Arudino UNO A4)
#define BT_TXpin 19 // Желтый провод -> Bluetooth RX (Arudino UNO A5)
// питание красный(VCC)-оранжевый(GND)
SoftwareSerial btSerial = SoftwareSerial (BT_RXpin, BT_TXpin); // RX, TX

// Символьные и строковые переменные для чтения буфера UART Bluetooth
char incomingByte;
String incomingString;
String GetBTCommand;

// Параметры
// Скорость движения робота = напряжение на моторах, диапазон 0-255
// Выбран 40% уровень напряжения питания моторов
byte MotorSpeed = 96;

// Длительность выполнения одного движения в мс
byte OneMovieTime = 3;

// Предельное прямое расстояние до объекта см
short LimitObjectDistance = 25;

// Текущее прямое расстояние до объекта см
short CurrentObjectDistance = 0;

// Режимные индикаторы
bool SonarMovie = false;
bool LineTracker = false;
bool BluetoothControl = false;

// Константы управления и сигнализации
#define FASTLEFT 0
#define FASTRIGHT 1
#define FORWARD 2
#define BACKWARD 3
#define RELEASE 4
#define TURNLEFT 5
#define TURNRIGHT 6
#define LeftTrackerAlarm 7
#define RightTrackerAlarm 8
#define AllTrackerAlarm 9
#define AllTrackerNormal 10

// Константы для расчета дистанции от ультразвукового сонара до объекта

// Скорость звука м/с в воздухе при t=20C
const float sound_speed = 343.1;

// Коэффициент преобразования скорости м/с -> см/мкс
// Грубый расчет по формуле 343 м/с = 0.0343 см/мкс (343.1*100/1000000) = 1/29.1 см/мкс, т.е. 29.1см за 1 мкс
// Результат умножается на 2, так как сигнал движется в обе стороны, от сонара до препятствия,
// и от препятствия до сонара.
// Итоговое значение коэффициента для исходной скорости звука 343,1 равно 58.2 см/мкс
const float k = (20000 / sound_speed);

// ================  переменные, процедуры и функции =====================


byte getTracker(byte RightValue, byte LeftValue) {
  // Проверка и возвращение состояния трекеров. HIGH - черный цвет, LOW - белый цвет

  // Нет сигналов с трекеров
  if (digitalRead(RightValue) == HIGH && digitalRead(LeftValue) == HIGH) return AllTrackerNormal;
  // Сигнал с левого трекера
  if (digitalRead(RightValue) == HIGH && digitalRead(LeftValue) == LOW) return LeftTrackerAlarm;
  // Сигнал с правого трекера
  if (digitalRead(RightValue) == LOW && digitalRead(LeftValue) == HIGH) return RightTrackerAlarm;
  // Сигнал с двух трекеров
  if (digitalRead(RightValue) == LOW && digitalRead(LeftValue) == LOW) return AllTrackerAlarm;
}

unsigned short getDistance(unsigned short maxDistance = 600) {
  // Функция определения расстояния от ультразвукового датчика до объекта
  // Возвращает -1, если превышен таймаут ожидания эхо
  // Необязательный параметр maxDistance задаёт максимальное измеряемое расстояние в сантиметрах
  unsigned long timeout = maxDistance * k;
  unsigned long previousMicros;
  unsigned int t;
  // Формирование импульса на входе Trig сонара длительностью 10мкс
  digitalWrite(pinTrig, LOW);
  delayMicroseconds(2);
  digitalWrite(pinTrig, HIGH);
  delayMicroseconds(10);
  digitalWrite(pinTrig, LOW);

  // Ожидание состояния HIGH на выводе Echo
  previousMicros = micros();
  while (!digitalRead(pinEcho)) {
    if (micros() - previousMicros > 10000) return -1;
  }

  // Измерение длительности ответного импульса на выводе Echo сонара
  previousMicros = micros();
  while (digitalRead(pinEcho)) {
    if (micros() - previousMicros > timeout) return -1;
  }
  t = micros() - previousMicros;
  return (t / k);
}

void SetPatch(byte WayDirection, unsigned int MoveTime)
// Управление входами драйвера моторов
//in1 - HIGH + in2 - LOW = левый мотор назад
//in2 - HIGH + in1 - LOW = левый мотор вперед
//in3 - HIGH + in4 - LOW = правый мотор назад
//in4 - HIGH + in3 - LOW = правый мотор вперед

{
  switch (WayDirection) {
    case FASTLEFT: {
        // быстрый разворот влево
        Serial.println("FAST LEFT");        
        analogWrite(MotorInput1, MotorSpeed);
        digitalWrite(MotorInput2, LOW);
        digitalWrite(MotorInput3, LOW);
        analogWrite(MotorInput4, MotorSpeed);
        delay(MoveTime);
        break;
      }
    case FASTRIGHT: {
        // быстрый разворот влево
        Serial.println("FAST RIGHT");
        digitalWrite(MotorInput1, LOW);
        analogWrite(MotorInput2, MotorSpeed);
        analogWrite(MotorInput3, MotorSpeed);
        digitalWrite(MotorInput4, LOW);        
        delay(MoveTime);
        break;
      }      
    case FORWARD: {
        // двжижение вперед
        Serial.println("Forward");
        digitalWrite(MotorInput1, LOW);
        analogWrite(MotorInput2, MotorSpeed);
        digitalWrite(MotorInput3, LOW);
        analogWrite(MotorInput4, MotorSpeed);
        delay(MoveTime);
        break;
      }
    case BACKWARD: {
        // двжижение назад
        Serial.println("Backward");
        analogWrite(MotorInput1, MotorSpeed);
        digitalWrite(MotorInput2, LOW);
        analogWrite(MotorInput3, MotorSpeed);
        digitalWrite(MotorInput4, LOW);
        delay(MoveTime);
        break;
      }
    case TURNLEFT: {
        // поворот влево
        Serial.println("Left");
        digitalWrite(MotorInput1, LOW);
        digitalWrite(MotorInput2, LOW);
        digitalWrite(MotorInput3, LOW);
        analogWrite(MotorInput4, MotorSpeed);
        delay(MoveTime);
        break;
      }
    case TURNRIGHT: {
        // поворот вправо
        Serial.println("Right");
        digitalWrite(MotorInput1, LOW);
        analogWrite(MotorInput2, MotorSpeed);
        digitalWrite(MotorInput3, LOW);
        digitalWrite(MotorInput4, LOW);
        delay(MoveTime);
        break;
      }
    case RELEASE: {
        // остановка
        Serial.println("Release");
        digitalWrite(MotorInput1, LOW);
        digitalWrite(MotorInput2, LOW);
        digitalWrite(MotorInput3, LOW);
        digitalWrite(MotorInput4, LOW);
        break;
      }
  }
}

// ================  подготовительная процедура, выполняется один раз в момент включения =====================

void setup() {
  // Определение алгоритма управления
  pinMode(ProgramSelect1, INPUT);
  pinMode(ProgramSelect2, INPUT);
    
  if (digitalRead(ProgramSelect1) == HIGH && digitalRead(ProgramSelect2) == LOW) {
    // Алгоритм движение по сонару и боковым датчикам препятствия, SW1:HIGH, SW2: LOW
    // Инициализация выводов сонара
    pinMode(pinTrig, OUTPUT);
    pinMode(pinEcho, INPUT);
    SonarMovie = true;
    LineTracker = false;
    BluetoothControl = false;
  }
  
  else if (digitalRead(ProgramSelect1) == LOW && digitalRead(ProgramSelect2) == LOW)  {
    // Алгоритм слежения за полосой, SW1:LOW, SW2: LOW
    // Инициализация выводов трекеров
    pinMode(RightTracker, INPUT);
    pinMode(LeftTracker, INPUT);
    SonarMovie = false;
    LineTracker = true;
    BluetoothControl = false;
  }
  
  else if (digitalRead(ProgramSelect1) == HIGH && digitalRead(ProgramSelect2) == HIGH)  {
    // Алгоритм управления по Bluetooth, SW1:HIGH, SW2: HIGH
    // Включение индикатора D13
    pinMode(LED_BUILTIN, OUTPUT);
    digitalWrite(LED_BUILTIN, HIGH);
    SonarMovie = false;
    LineTracker = false;
    BluetoothControl = true;
  } 
  
  else {
    // пропуск комбинации ProgramSelect1 == LOW и ProgramSelect2 == HIGH
  }
  
  // Инициализация выводов драйвера мотора
  pinMode(MotorInput1, OUTPUT);
  pinMode(MotorInput2, OUTPUT);
  pinMode(MotorInput3, OUTPUT);
  pinMode(MotorInput4, OUTPUT);

  // Запуск аппаратного канала последовательной связи Serial0 UART
  // для вывода отладочной информации через порт USB
  Serial.begin (9600);
  
  // Запуск программного канала последовательной связи Serial UART для модуля Bluetooth HC-05
  pinMode(BT_RXpin, INPUT);
  pinMode(BT_TXpin, OUTPUT);
  btSerial.begin (9600);
}

// ================  основной цикл =====================

void loop() {
  // Заголовок отладочной информации о направлении движения
  Serial.print("Move - ");
  // Выполнение алгоритмов управления и движения в соответствии с выбранной программой
  // Алгоритм движения по полосе
  if (LineTracker == true && SonarMovie == false && BluetoothControl == false) {
    switch (getTracker(RightTracker, LeftTracker)) {
      // сигнал от левого датчика, поворачиваем вправо
      case LeftTrackerAlarm: {
          SetPatch(TURNRIGHT, OneMovieTime);
          break;
        }
      // сигнал от правого датчика, поворачиваем влево
      case RightTrackerAlarm: {
          SetPatch(TURNLEFT, OneMovieTime);
          break;
        }
      // сигнал от левого и правого датчика, поворачиваем влево
      case AllTrackerAlarm: {
          SetPatch(FASTLEFT, OneMovieTime);
          break;
        }
      // нет сигналов от датчиков, движемся прямо
      case AllTrackerNormal: {
          SetPatch(FORWARD, OneMovieTime);
          break;
        }
    }
  }
  
  // Алгоритм движения по сонару и боковым датчикам препятствий
  if (LineTracker == false && SonarMovie == true && BluetoothControl == false) {
    // Измирение дистанции до объекта по сонару, датчики настраиваются вручную
    CurrentObjectDistance = getDistance();

    // Сигнализируют оба датчика препятствия, быстрый поворот вправо
    if (getTracker(RightObstacle, LeftObstacle) == AllTrackerAlarm) {
      SetPatch(FASTRIGHT, OneMovieTime);
      
    // Левый датчики препятствия сигнализирует, быстрый поворот вправо
    } else if (getTracker(RightObstacle, LeftObstacle) == LeftTrackerAlarm) {
      SetPatch(FASTRIGHT, OneMovieTime);
      
    // Правый датчики препятствия сигнализирует, быстрый поворот влево
    } else if (getTracker(RightObstacle, LeftObstacle) == RightTrackerAlarm) {
      SetPatch(FASTLEFT, OneMovieTime);
      
    // Текущая дистанция до препятствия по сонару меньше пороговой, боковые препятствия не регистрируются датчиками
    } else if (CurrentObjectDistance < LimitObjectDistance && getTracker(RightObstacle, LeftObstacle) == AllTrackerNormal) {
      SetPatch(FASTRIGHT, OneMovieTime);

    // Движение вперед
    } else {   
      SetPatch(FORWARD, OneMovieTime);
    }

    // Вывод отладочной информации о текущем расcтоянии от сонара до объекта
    Serial.print(LimitObjectDistance);
    Serial.print(" cm limit, current distance : ");
    Serial.println(CurrentObjectDistance);
  }
  
  // Алгоритм движения с управлением из Bluetooth-приложения
  if (LineTracker == false && SonarMovie == false && BluetoothControl == true) {

    // В буфере появились новые данные
    if (btSerial.available() > 0) {
      // Посимвольное считывание до ключевого знака "0"
      incomingByte = btSerial.read();
      // Составление команды из нескольких символов
      incomingString += incomingByte;
      // Считываение ключевого знака "0" завершает формирование команды
      if (incomingByte == '0') {
        GetBTCommand = incomingString;
        Serial.println(GetBTCommand);
        incomingString = "";
      }
    } 
    // Выполнение текущей каманды, пока в буфере нет новых данных
    if (GetBTCommand == "FWD0") {
      SetPatch(FORWARD, OneMovieTime);
    } else if (GetBTCommand == "BWD0") {
      SetPatch(BACKWARD, OneMovieTime);
    } else if (GetBTCommand == "RGT0") {
      SetPatch(TURNRIGHT, OneMovieTime);
    } else if (GetBTCommand == "LFT0") {
      SetPatch(TURNLEFT, OneMovieTime);
    } else if (GetBTCommand == "RLS0") {
      SetPatch(RELEASE, OneMovieTime);
    } else {
      // Сброс строки при несовпадении с командой управления
      GetBTCommand = "";
    }
  }
}
